local ffi = require "ffi"
local ffi_new = ffi.new
local ffi_str = ffi.string
local C = ffi.C

ffi.cdef[[
int RAND_bytes(unsigned char *buf, int num);
int RAND_pseudo_bytes(unsigned char *buf, int num);
]]

assert(ffi.load("libssh"), "can't read libssh!")

local _M = {}

_M.random = {
	bytes = function(len, strong)
		local buf = ffi_new("char[?]", len)
		if strong then
			if C.RAND_bytes(buf, len) == 0 then
				return nil
			end
		else
			C.RAND_pseudo_bytes(buf,len)
		end

		return ffi_str(buf, len)
	end
}

_M.string = {
	to_hex = function(str)
		return (str:gsub('.', function (c)
			return string.format('%x', string.byte(c))
		end))
	end
}

return _M